@@ -7,7 +7,7 @@ import sys, os, skytools
 from installer import *
 
 # support set() on 2.3
-if 'set' not in dir(__builtins__):
+if 'set' not in __builtins__:
     from sets import Set as set
 
 __all__ = ['ProviderSetup', 'SubscriberSetup']
@@ -56,11 +56,19 @@ class CommonSetup(skytools.DBScript):
         self.pgq_queue_name = self.cf.get("pgq_queue_name")
         self.consumer_id = self.cf.get("pgq_consumer_id", self.job_name)
         self.fake = self.cf.getint('fake', 0)
+        self.lock_timeout = self.cf.getfloat('lock_timeout', 10)
 
         if len(self.args) < 3:
             self.log.error("need subcommand")
             sys.exit(1)
 
+    def set_lock_timeout(self, curs):
+        ms = int(1000 * self.lock_timeout)
+        if ms > 0:
+            q = "SET LOCAL statement_timeout = %d" % ms
+            self.log.debug(q)
+            curs.execute(q)
+
     def run(self):
         self.admin()
 
@@ -93,10 +101,12 @@ class CommonSetup(skytools.DBScript):
         return res
 
     def get_all_seqs(self, curs):
-        q = """SELECT n.nspname || '.'|| c.relname
+        q = """SELECT n.nspname || '.' || c.relname
                  from pg_class c, pg_namespace n
                 where n.oid = c.relnamespace 
                   and c.relkind = 'S'
+                  and n.nspname not in ('pgq', 'londiste', 'pgq_node')
+                  and n.nspname !~ '^pg_temp_.*'
                 order by 1"""
         curs.execute(q)
         res = []
@@ -186,9 +196,18 @@ class ProviderSetup(CommonSetup):
         if not seq_list and self.options.all:
             seq_list = self.provider_get_all_seqs()
 
-        for seq in self.args[3:]:
+        cur_list = self.get_provider_seqs()
+        gotnew = False
+        for seq in seq_list:
+            seq = skytools.fq_name(seq)
+            if seq in cur_list:
+                self.log.info('Seq %s already subscribed' % seq)
+                continue
+            gotnew = True
             self.provider_add_seq(seq)
-        self.provider_notify_change()
+
+        #if gotnew:
+        #    self.provider_notify_change()
 
     def provider_remove_seq_list(self, seq_list):
         if not seq_list and self.options.all:
@@ -239,7 +258,7 @@ class ProviderSetup(CommonSetup):
                     self.provider_add_table(tbl)
                 else:
                     self.log.info("Table %s already added" % tbl)
-        self.provider_notify_change()
+        #self.provider_notify_change()
 
     def provider_remove_tables(self, table_list):
         self.check_provider_queue()
@@ -262,6 +281,31 @@ class ProviderSetup(CommonSetup):
         q = "select londiste.provider_add_table(%s, %s)"
         self.exec_provider(q, [self.pgq_queue_name, tbl])
 
+        # detect dangerous triggers
+        q = """
+            select tg.trigger_name
+                from londiste.provider_table tbl,
+                     information_schema.triggers tg
+                where tbl.queue_name = %s
+                  and tbl.table_name = %s
+                  and tg.event_object_schema = %s
+                  and tg.event_object_table = %s
+                  and tg.condition_timing = 'AFTER'
+                  and tg.trigger_name != tbl.trigger_name
+                  and tg.trigger_name < tbl.trigger_name
+                  and substring(tg.trigger_name from 1 for 10) != '_londiste_'
+                  and substring(tg.trigger_name from char_length(tg.trigger_name) - 6) != '_logger'
+            """
+        sname, tname = skytools.fq_name_parts(tbl)
+        src_db = self.get_database('provider_db')
+        src_curs = src_db.cursor()
+        src_curs.execute(q, [self.pgq_queue_name, tbl, sname, tname])
+        for r in src_curs.fetchall():
+            self.log.warning("Table %s has AFTER trigger '%s' which runs before Londiste trigger.  "\
+                             "If it modifies data, then events will appear in queue in wrong order." % (
+                                 tbl, r[0]))
+        src_db.commit()
+
     def provider_remove_table(self, tbl):
         q = "select londiste.provider_remove_table(%s, %s)"
         self.exec_provider(q, [self.pgq_queue_name, tbl])
@@ -290,6 +334,8 @@ class ProviderSetup(CommonSetup):
         src_db = self.get_database('provider_db')
         src_curs = src_db.cursor()
 
+        self.set_lock_timeout(src_curs)
+
         src_curs.execute(sql, args)
 
         if self.fake:
