@@ -60,7 +60,7 @@ class CopyTable(Replicator):
                 raise Exception('Column %s does not exist on dest side' % c)
 
         # drop unnecessary stuff
-        objs = T_CONSTRAINT | T_INDEX | T_RULE
+        objs = T_CONSTRAINT | T_INDEX | T_RULE | T_PARENT
         dst_struct.drop(dst_curs, objs, log = self.log)
 
         # do truncate & copy
@@ -108,7 +108,13 @@ class CopyTable(Replicator):
             self.log.info("%s: skipping truncate" % tablename)
         else:
             self.log.info("%s: truncating" % tablename)
-            dstcurs.execute("truncate " + skytools.quote_fqident(tablename))
+            # truncate behaviour changed in 8.4
+            dstcurs.execute("show server_version_num")
+            pgver = int(dstcurs.fetchone()[0])
+            if pgver >= 80400:
+                dstcurs.execute("truncate only " + skytools.quote_fqident(tablename))
+            else:
+                dstcurs.execute("truncate " + skytools.quote_fqident(tablename))
 
         # do copy
         self.log.info("%s: start copy" % tablename)
