@@ -37,18 +37,10 @@ Switches:
   -n                 no action, just print commands
 """
 
-"""
-Additional features:
- * Simplified install. Master "setup" command should setup slave directories.
- * Add support for multiple targets on master.
- * WAL purge does not correctly purge old WAL-s if timelines are involved. The first
- useful WAL name is obtained from backup_label, WAL-s in the same timeline that are older 
- than first useful WAL are removed. 
- * Always copy the directory on "restore" add a special "--move" option.
-"""
+import os, sys, re, signal, time, traceback
+import errno, glob, ConfigParser, shutil, subprocess
 
-import os, sys, skytools, re, signal, time, traceback
-import errno, glob, ConfigParser, shutil
+import skytools
 
 MASTER = 1
 SLAVE = 0
@@ -114,19 +106,27 @@ class WalChunk:
 class PgControlData:
     """Contents of pg_controldata"""
 
-    def __init__(self, slave_bin, slave_data, findRestartPoint):
+    def __init__(self, bin_dir, data_dir, findRestartPoint):
         """Collect last checkpoint information from pg_controldata output"""
         self.xlogid = None
         self.xrecoff = None
         self.timeline = None
         self.wal_size = None
         self.wal_name = None
-        self.is_valid = False
+        self.cluster_state = None
+        self.is_shutdown = False
         self.pg_version = 0
-        matches = 0
-        pg_controldata = "%s %s" % (os.path.join(slave_bin, "pg_controldata"), slave_data)
+        self.is_valid = False
 
-        for line in os.popen(pg_controldata, "r"):
+        try:
+            pg_controldata = os.path.join(bin_dir, "pg_controldata")
+            pipe = subprocess.Popen([ pg_controldata, data_dir ], stdout=subprocess.PIPE)
+        except OSError:
+            # don't complain if we cannot execute it
+            return
+
+        matches = 0
+        for line in pipe.stdout.readlines():
             if findRestartPoint:
                 m = re.match("^Latest checkpoint's REDO location:\s+([0-9A-F]+)/([0-9A-F]+)", line)
             else:
@@ -147,8 +147,14 @@ class PgControlData:
             if m:
                 matches += 1
                 self.pg_version = int(m.group(1))
+            m = re.match("^Database cluster state:\s+(.*$)", line)
+            if m:
+                matches += 1
+                self.cluster_state = m.group(1)
+                self.is_shutdown = (self.cluster_state == "shut down")
 
-        if matches == 4:
+        # ran successfully and we got our needed matches
+        if pipe.wait() == 0 and matches == 5:
             self.wal_name = "%08X%08X%08X" % \
                 (self.timeline, self.xlogid, self.xrecoff / self.wal_size)
             self.is_valid = True
@@ -198,29 +204,40 @@ class PostgresConfiguration:
             return "off"
         return None
 
+    def wal_level(self):
+        """Return value for specified parameter"""
+        # see if explicitly set
+        m = re.search("^\s*wal_level\s*=\s*'?([a-z_]+)'?\s*#?.*$", self.cf_buf, re.M | re.I)
+        if m:
+            return m.group(1)
+        # also, it could be commented out as initdb leaves it
+        # it'd probably be best to check from the database ...
+        m = re.search("^#wal_level\s*=.*$", self.cf_buf, re.M | re.I)
+        if m:
+            return "minimal"
+        return None
+
     def modify(self, cf_params):
         """Change the configuration parameters supplied in cf_params"""
 
         for (param, value) in cf_params.iteritems():
-            r_active = re.compile("^[ ]*%s[ ]*=[ ]*'(.*)'.*$" % param, re.M)
-            r_disabled = re.compile("^.*%s.*$" % param, re.M)
+            r_active = re.compile("^\s*%s\s*=\s*([^\s#]*).*$" % param, re.M)
+            r_disabled = re.compile("^\s*#\s*%s\s*=.*$" % param, re.M)
 
             cf_full = "%s = '%s'" % (param, value)
 
             m = r_active.search(self.cf_buf)
             if m:
                 old_val = m.group(1)
-                if old_val == value:
-                    self.log.debug("parameter %s already set to '%s'" % (param, value))
-                else:
-                    self.log.debug("found parameter %s with value '%s'" % (param, old_val))
-                    self.cf_buf = "%s%s%s" % (self.cf_buf[:m.start()], cf_full, self.cf_buf[m.end():])
+                self.log.debug("found parameter %s with value '%s'" % (param, old_val))
+                self.cf_buf = "%s%s%s" % (self.cf_buf[:m.start()], cf_full, self.cf_buf[m.end():])
             else:
                 m = r_disabled.search(self.cf_buf)
                 if m:
                     self.log.debug("found disabled parameter %s" % param)
                     self.cf_buf = "%s\n%s%s" % (self.cf_buf[:m.end()], cf_full, self.cf_buf[m.end():])
                 else:
+                    # not found, append to the end
                     self.log.debug("found no value")
                     self.cf_buf = "%s\n%s\n\n" % (self.cf_buf, cf_full)
 
@@ -252,6 +269,9 @@ class WalMgr(skytools.DBScript):
 
     def __init__(self, args):
 
+        if len(args) == 1 and args[0] == '--version':
+           skytools.DBScript.__init__(self, 'wal-master', args)
+
         if len(args) < 2:
             # need at least config file and command
             usage(1)
@@ -385,6 +405,45 @@ class WalMgr(skytools.DBScript):
             self.pg_stop_backup()
             sys.exit(1)
 
+    def rsync_log_directory(self, source_dir, dst_loc):
+        """rsync a pg_log or pg_xlog directory - ignore most of the 
+        directory contents, and pay attention to symlinks
+        """
+        keep_symlinks = self.cf.getint("keep_symlinks", 1)
+
+        subdir = os.path.basename(source_dir)
+        if not os.path.exists(source_dir):
+            self.log.info("%s does not exist, skipping" % subdir)
+            return
+
+        cmdline = []
+
+        # if this is a symlink, copy it's target first
+        if os.path.islink(source_dir) and keep_symlinks:
+            self.log.info('%s is a symlink, attempting to create link target' % subdir)
+
+            # expand the link
+            link = os.readlink(source_dir)
+            if not link.startswith("/"):
+                link = os.path.join(os.getcwd(), link)
+            link_target = os.path.join(link, "")
+
+            remote_target = "%s:%s" % (self.slave_host(), link_target)
+            options = [ "--include=archive_status", "--exclude=/**" ]
+            if self.exec_rsync( options + [ link_target, remote_target ]):
+                # unable to create the link target, just convert the links
+                # to directories in PGDATA
+                self.log.warning('Unable to create symlinked %s on target, copying' % subdir)
+                cmdline += [ "--copy-unsafe-links" ]
+
+        cmdline += [ "--exclude=pg_log/*" ]
+        cmdline += [ "--exclude=pg_xlog/archive_status/*" ]
+        cmdline += [ "--include=pg_xlog/archive_status" ]
+        cmdline += [ "--exclude=pg_xlog/*" ]
+
+        self.exec_big_rsync(cmdline + [ source_dir, dst_loc ])
+
+
     def exec_cmd(self, cmdline):
         cmd = "' '".join(cmdline)
         self.log.debug("Execute cmd: '%s'" % (cmd))
@@ -438,8 +497,6 @@ class WalMgr(skytools.DBScript):
         except:
             self.log.fatal("Cannot write to %s" % fn)
 
-    def authfile_name(self):
-        return os.path.join(self.cf.get("master_data"), os.path.join("global", "pg_auth"))
 
     def master_stop(self):
         """Deconfigure archiving, attempt to stop syncdaemon"""
@@ -470,38 +527,61 @@ class WalMgr(skytools.DBScript):
         """Turn the archiving on or off"""
 
         cf = PostgresConfiguration(self, self.cf.get("master_config"))
-        archive_mode = cf.archive_mode()
+        curr_archive_mode = cf.archive_mode()
+        curr_wal_level = cf.wal_level()
+        need_restart_warning = False
 
         if enable_archiving:
             # enable archiving
             cf_file = os.path.abspath(self.cf.filename)
+
             xarchive = "%s %s %s" % (self.script, cf_file, "xarchive %p %f")
             cf_params = { "archive_command": xarchive }
 
-            if archive_mode:
+            if curr_archive_mode is not None:
                 # archive mode specified in config, turn it on
                 self.log.debug("found 'archive_mode' in config -- enabling it")
                 cf_params["archive_mode"] = "on"
 
-                if archive_mode.lower() not in ('1', 'on', 'true') and not can_restart:
-                    self.log.warning("database must be restarted to enable archiving")
-            else:
-                self.log.debug("seems that archive_mode is not set, ignoring it.")
+                if curr_archive_mode.lower() not in ('1', 'on', 'true') and not can_restart:
+                    need_restart_warning = True
+
+            if curr_wal_level is not None and curr_wal_level != 'hot_standby':
+                # wal level set in config, enable it
+                wal_level = self.cf.getboolean("hot_standby", False) and "hot_standby" or "archive"
+
+                self.log.debug("found 'wal_level' in config -- setting to '%s'" % wal_level)
+                cf_params["wal_level"] = wal_level
+
+                if curr_wal_level not in ("archive", "hot_standby") and not can_restart:
+                    need_restart_warning = True
+
+            if need_restart_warning:
+                self.log.warning("database must be restarted to enable archiving")
 
         else:
             # disable archiving
-            if not archive_mode:
-                # archive_mode not set, just reset archive_command and its done
-                self.log.debug("archive_mode not set in config, leaving as is")
-                cf_params = { "archive_command": "" }
-            elif can_restart:
-                # restart possible, turn off archiving, reset archive_command
-                cf_params = { "archive_mode": "off", "archive_command": "" }
-            else:
-                # not possible to change archive_mode (requires restart), so we
-                # just set the archive_command to /bin/true to avoid WAL pileup.
-                self.log.debug("master_restart_cmd not specified, leaving archive_mode as is")
-                cf_params = { "archive_command": "/bin/true" }
+
+            cf_params = {
+                "archive_mode": "off",
+                "wal_level": "minimal",
+                "archive_command": ""
+            }
+
+            if curr_archive_mode is None:
+                self.log.info("archive_mode not set")
+                del(cf_params['archive_mode'])
+
+            if curr_wal_level is None:
+                self.log.info("wal_level not set")
+                del(cf_params['wal_level'])
+
+            if not can_restart:
+                # not possible to change archive_mode or wal_level (requires restart),
+                # so we just set the archive_command to /bin/true to avoid WAL pileup.
+                self.log.warning("database must be restarted to disable archiving")
+                self.log.info("Setting archive_command to /bin/true to avoid WAL pileup")
+                cf_params['archive_command'] = '/bin/true'
 
         self.log.debug("modifying configuration: %s" % cf_params)
 
@@ -509,15 +589,15 @@ class WalMgr(skytools.DBScript):
         cf.write()
 
     def slave_deconfigure_archiving(self, cf_file):
-        """Turn the archiving off for the slave"""
+        """Disable archiving for the slave. This is done by setting
+        archive_command to a trivial command, so that archiving can be
+        re-enabled without restarting postgres. Needed when slave is
+        booted with postgresql.conf from master."""
 
         self.log.debug("Disable archiving in %s" % cf_file)
 
         cf = PostgresConfiguration(self, cf_file)
-
-        cf_params = { "archive_command": "" }
-        if cf.archive_mode():
-            cf_params["archive_mode"] = "off"
+        cf_params = { "archive_command": "/bin/true" }
 
         self.log.debug("modifying configuration: %s" % cf_params)
         cf.modify(cf_params)
@@ -534,6 +614,15 @@ class WalMgr(skytools.DBScript):
             cmdline = ["ssh", "-nT", host, "mkdir", "-p", path]
             self.exec_cmd(cmdline)
 
+    def slave_host(self):
+        """Extract the slave hostname"""
+        try:
+            slave = self.cf.get("slave")
+            host, path = slave.split(":", 1)
+        except:
+            raise Exception("invalid value for 'slave' in %s" % self.cfgfile)
+        return host
+
     def remote_walmgr(self, command, stdin_disabled = True):
         """Pass a command to slave WalManager"""
 
@@ -665,7 +754,7 @@ class WalMgr(skytools.DBScript):
                     "--exclude", "*.conf",
                     "--exclude", "pg_xlog",
                     "--exclude", "pg_tblspc",
-                    "--exclude", "pg_log/*",
+                    "--exclude", "pg_log",
                     "--copy-unsafe-links",
                     ".", dst_loc]
             self.exec_big_rsync(cmdline)
@@ -695,14 +784,10 @@ class WalMgr(skytools.DBScript):
                     cmdline = [ "--delete", "--exclude", ".*", "--copy-unsafe-links", ".", dstfn]
                     self.exec_big_rsync(cmdline)
 
-            # copy pg_xlog
-            self.chdir(data_dir)
-            cmdline = [
-                "--exclude", "*.done",
-                "--exclude", "*.backup",
-                "--copy-unsafe-links",
-                "--delete", "pg_xlog", dst_loc]
-            self.exec_big_rsync(cmdline)
+            # copy the pg_log and pg_xlog directories, these may be 
+            # symlinked to nonstandard location, so pay attention
+            self.rsync_log_directory(os.path.join(data_dir, "pg_log"),  dst_loc)
+            self.rsync_log_directory(os.path.join(data_dir, "pg_xlog"), dst_loc)
 
             # copy config files
             conf_dst_loc = self.cf.get("config_backup", "")
@@ -860,7 +945,7 @@ STOP TIME: %(stop_time)s
             dst_loc += "/"
 
         # copy data
-        self.exec_rsync([ srcpath, self.authfile_name(), dst_loc ], True)
+        self.exec_rsync([ srcpath, dst_loc ], True)
 
         self.log.debug("%s: done", srcname)
         end_time = time.time()
@@ -973,17 +1058,10 @@ STOP TIME: %(stop_time)s
         use_xlog_functions = self.cf.getint("use_xlog_functions", False)
         data_dir = self.cf.get("master_data")
         xlog_dir = os.path.join(data_dir, "pg_xlog")
+        master_bin = self.cf.get("master_bin", "")
 
-        auth_loc = os.path.join(self.cf.get("completed_wals"), "")
         dst_loc = os.path.join(self.cf.get("partial_wals"), "")
 
-        # sync the auth file
-        if not daemon_mode:
-            # avoid the extra rsync in daemon mode - the file is fairly static, so we 
-            # don't need to sync it every N seconds.
-            if self.exec_rsync([self.authfile_name(), auth_loc]) != 0:
-                self.log.warning('Cannot sync auth file')
-
         db = None
         if use_xlog_functions:
             try:
@@ -1023,6 +1101,22 @@ STOP TIME: %(stop_time)s
             else:
                 self.log.info("last complete not found, copying all")
 
+            # obtain the last checkpoint wal name, this can be used for
+            # limiting the amount of WAL files to copy if the database
+            # has been cleanly shut down
+            ctl = PgControlData(master_bin, data_dir, False)
+            checkpoint_wal = None
+            if ctl.is_valid:
+                if not ctl.is_shutdown:
+                    # cannot rely on the checkpoint wal, should use some other method
+                    self.log.info("Database state is not 'shut down', copying all")
+                else:
+                    # ok, the database is shut down, we can use last checkpoint wal
+                    checkpoint_wal = ctl.wal_name
+                    self.log.info("last checkpoint wal: %s" % checkpoint_wal)
+            else:
+                self.log.info("Unable to obtain control file information, copying all")
+
             for fn in files:
                 # check if interesting file
                 if len(fn) < 10:
@@ -1031,7 +1125,7 @@ STOP TIME: %(stop_time)s
                     continue
                 if fn.find(".") > 0:
                     continue
-                # check if to old
+                # check if too old
                 if last:
                     dot = last.find(".")
                     if dot > 0:
@@ -1041,6 +1135,9 @@ STOP TIME: %(stop_time)s
                     else:
                         if fn <= last:
                             continue
+                # check if too new
+                if checkpoint_wal and fn > checkpoint_wal:
+                    continue
 
                 # got interesting WAL
                 xlog = os.path.join(xlog_dir, fn)
@@ -1168,12 +1265,6 @@ STOP TIME: %(stop_time)s
             self.log.debug("%s: copy done, cleanup" % srcname)
             self.slave_cleanup(lstname)
 
-        if os.path.isfile(partfile) and not srcfile == partfile:
-            # Remove any partial files after restore. Only leave the partial if
-            # it is actually used in recovery.
-            self.log.debug("%s: removing partial not anymore needed for recovery." % partfile)
-            os.remove(partfile)
-
         # create a PROGRESS file to notify that postgres is processing the WAL
         open(prgrfile, "w").write("1")
 
@@ -1186,7 +1277,7 @@ STOP TIME: %(stop_time)s
 
         If setname is specified, the contents of that backup set directory are 
         restored instead of "full_backup". Also copy is used instead of rename to 
-        restore the directory.
+        restore the directory (unless a pg_xlog directory has been specified).
 
         Restore to altdst if specified. Complain if it exists.
         """
@@ -1245,9 +1336,15 @@ STOP TIME: %(stop_time)s
             # nothing to back up
             createbackup = False
 
-        if not setname and os.path.isdir(data_dir):
-            # compatibility mode - restore without a set name and data directory 
-            # already exists. Move it out of the way.
+        # see if we have to make a backup of the data directory 
+        backup_datadir = self.cf.getboolean('backup_datadir', True)
+
+        if os.path.isdir(data_dir) and not backup_datadir:
+            self.log.warning('backup_datadir is disabled, deleting old data dir')
+            shutil.rmtree(data_dir)
+
+        if not setname and os.path.isdir(data_dir) and backup_datadir:
+            # compatibility mode - restore without a set name and data directory exists
             self.log.warning("Data directory already exists, moving it out of the way.")
             createbackup = True
 
@@ -1259,16 +1356,38 @@ STOP TIME: %(stop_time)s
 
         # move new data, copy if setname specified
         self.log.info("%s %s to %s" % (setname and "Copy" or "Move", full_dir, data_dir))
+
+        if self.cf.get('slave_pg_xlog', ''):
+            link_xlog_dir = True
+            exclude_pg_xlog = '--exclude=pg_xlog'
+        else:
+            link_xlog_dir = False
+            exclude_pg_xlog = ''
+
         if not self.not_really:
-            if not setname:
+            if not setname and not link_xlog_dir:
                 os.rename(full_dir, data_dir)
             else:
-                self.exec_rsync(["--delete", "--no-relative", "--exclude=pg_xlog/*",
-                    os.path.join(full_dir,""), data_dir], True)
-                if self.wtype == MASTER and createbackup and os.path.isdir(bak):
+                rsync_args=["--delete", "--no-relative", "--exclude=pg_xlog/*"]
+                if exclude_pg_xlog:
+                    rsync_args.append(exclude_pg_xlog)
+                rsync_args += [os.path.join(full_dir, ""), data_dir]
+
+                self.exec_rsync(rsync_args, True)
+
+                if link_xlog_dir:
+                   os.symlink(self.cf.get('slave_pg_xlog'), "%s/pg_xlog" % data_dir)
+
+                if (self.wtype == MASTER and createbackup and os.path.isdir(bak)):
                     # restore original xlog files to data_dir/pg_xlog   
-                    # symlinked directories are dereferences
-                    self.exec_cmd(["cp", "-rL", "%s/pg_xlog" % bak, data_dir])
+                    # symlinked directories are dereferenced
+                    self.exec_cmd(["cp", "-rL", "%s/pg_xlog/" % full_dir, "%s/pg_xlog" % data_dir ])
+                else:
+                    # create an archive_status directory
+                    xlog_dir = os.path.join(data_dir, "pg_xlog")
+                    archive_path = os.path.join(xlog_dir, "archive_status")
+                    if not os.path.exists(archive_path):
+                        os.mkdir(archive_path, 0700)
         else:
             data_dir = full_dir
 
@@ -1404,8 +1523,6 @@ restore_command = '%s %s %s'
         srcdir = self.cf.get("completed_wals")
         datadir = self.cf.get("slave_data")
         stopfile = os.path.join(srcdir, "STOP")
-        src_authfile = os.path.join(srcdir, "pg_auth")
-        dst_authfile = os.path.join(datadir, os.path.join("global", "pg_auth"))
 
         if self.not_really:
             self.log.info("Writing STOP file: %s" % stopfile)
@@ -1413,13 +1530,6 @@ restore_command = '%s %s %s'
             open(stopfile, "w").write("1")
         self.log.info("Stopping recovery mode")
 
-        if os.path.isfile(src_authfile):
-            self.log.debug("Using pg_auth file from master.")
-            try:
-                os.rename(dst_authfile, "%s.old" % dst_authfile)
-                self.exec_cmd(["cp", src_authfile, dst_authfile])
-            except Exception, e:
-                self.log.warning("Unable to restore pg_auth file: %s" % e)
 
     def slave_pause(self, waitcomplete=0):
         """Pause the WAL apply, wait until last file applied if needed"""
@@ -1629,14 +1739,14 @@ restore_command = '%s %s %s'
         partial_wals = self.cf.get("partial_wals")
 
         self.log.debug("cleaning completed wals before %s" % last_applied)
-        last = self.del_wals(completed_wals, last_applied)
-        if last:
-            if os.path.isdir(partial_wals):
-                self.log.debug("cleaning partial wals before %s" % last)
-                self.del_wals(partial_wals, last)
-            else:
-                self.log.warning("partial_wals dir does not exist: %s"
-                              % partial_wals)
+        self.del_wals(completed_wals, last_applied)
+
+        if os.path.isdir(partial_wals):
+            self.log.debug("cleaning partial wals before %s" % last_applied)
+            self.del_wals(partial_wals, last_applied)
+        else:
+            self.log.warning("partial_wals dir does not exist: %s" % partial_wals)
+
         self.log.debug("cleaning done")
 
     def del_wals(self, path, last):
